from abc import abstractmethod, ABC


class Entity(ABC):

    @abstractmethod
    def _key(self):
        pass

    def __eq__(self, other):
        return other is not None and self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __str__(self):
        return str(self._key())

    def __repr__(self):
        return self.__str__()



from pathlib import Path

import os

from modules.exceptions import AbsentUserConfigException


class UserConfig:
    CONFIG_FILE_NAME = '.az'
    API_KEY_CONFIG_NAME = 'key'
    INPUT_FILE_CONFIG_NAME = 'input_file'
    SEPARATOR = '='

    def __init__(self, key_arg, in_arg):
        self.key, self.in_file, self._key_local, self._in_local, self._key_global, self._in_global = None, None, None, None, None, None
        self._key_arg = key_arg
        self._in_arg = in_arg
        self._err_msg = ''
        self._read_local_config()
        self._read_global_config()
        self._assign_key()
        self._assign_in_file()

        if self._err_msg:
            raise AbsentUserConfigException(self._err_msg)

    def _read_local_config(self):
        self._key_local, self._in_local = self._read_user_config(self.CONFIG_FILE_NAME)

    def _read_global_config(self):
        home_dir = Path.home()
        global_config = os.path.join(home_dir, self.CONFIG_FILE_NAME)
        self._key_global, self._in_global = self._read_user_config(global_config)

    def _read_user_config(self, path):
        api_key, input_file = None, None
        if os.path.exists(path):
            with open(path) as config:
                for line in config:
                    if self.SEPARATOR in line:
                        key, value = line.split(self.SEPARATOR)
                        if key == self.API_KEY_CONFIG_NAME:
                            api_key = value.strip()
                        elif key == self.INPUT_FILE_CONFIG_NAME:
                            input_file = value.strip()
        return api_key, input_file

    def _assign_key(self):
        self.key = self._key_arg or self._key_local or self._key_global
        if not self.key:
            self._err_msg += "Key is not defined. Please, define  configuration parameter 'key' in local or global config"

    def _assign_in_file(self):
        self.in_file = self._in_arg or self._in_local or self._in_global
        if not self.in_file:
            self._err_msg += 'Input file is not defined.'





class Arguments(object):
    def __init__(self, number=None, dexdate=None, apksize=None, vtdetection=None, markets=None, pkgname=None, metadata=None, sha256=None, sha1=None, md5=None, key=None, input_file=None):
        self.number = number
        self.dexdate = dexdate
        self.apksize = apksize
        self.vtdetection = vtdetection
        self.markets = markets
        self.pkgname = pkgname
        self.metadata = metadata
        self.sha256 = sha256
        self.sha1 = sha1
        self.md5 = md5
        self.key = key
        self.input_file = input_file



from modules.entities.entity import Entity


class MarketsSet(set):
    def __init__(self, markets):
        super().__init__(markets.split('|'))

    def __str__(self):
        return '|'.join(sorted(self))


class Apk(Entity):
    def __init__(self, sha256=None, sha1=None, md5=None, dex_date=None, apk_size=None, pkg_name=None, vercode=None, vt_detection=None, vt_scan_date=None, dex_size=None, markets=None):
        self.sha256 = sha256
        self.sha1 = sha1
        self.vercode = int(vercode) if vercode else None
        self.md5 = md5
        self.dex_date = dex_date  # for performance reasons not parsing date here
        self.apk_size = int(apk_size) if apk_size else None
        self.pkg_name = pkg_name
        self.vt_detection = int(vt_detection) if vt_detection else None
        self.vt_scan_date = vt_scan_date
        self.dex_size = int(dex_size) if dex_size else None
        self.markets = MarketsSet(markets) if markets else None

    def _key(self):
        return self.sha256



import logging

from modules.entities.entity import Entity
from modules.services.apk_evaluator import ApkEvaluator


class FilteredDataset(Entity):
    def __init__(self, source=None, criteria=None):
        self.source = source
        self.criteria = criteria
        self.apk_evaluator = ApkEvaluator(criteria)
        self.length = 0
        logging.info(f'CREATE DATASET BY CRITERIA: {self.apk_evaluator.criteria}')

    def _key(self):
        return self.criteria

    def __iter__(self):
        return self

    def __next__(self):
        try:
            apk = next(self.source)
            logging.debug(f'EVALUATE {apk.pkg_name}')
            while not self.apk_evaluator.satisfies(apk):
                logging.debug(f'EVALUATE {apk.pkg_name}')
                apk = next(self.source)
            self.length += 1
            return apk
        except StopIteration:
            logging.info(f'DATASET HAS BEEN CREATED. {self.length} APKS SATISFY CRITERIA')
            raise StopIteration



from modules.exceptions import ValidationException
from modules.exceptions import NoArgsException


class Validator:
    def __init__(self, args):
        self.args = args

    def validate(self):
        self._check_at_least_one_arg_provided()
        self._check_number_is_positive()
        self._check_only_one_hash_is_used()

    def _check_at_least_one_arg_provided(self):
        if all(value is None for value in vars(self.args).values()):
            raise NoArgsException()

    def _check_only_one_hash_is_used(self):
        if (self.args.sha256 and self.args.sha1) or (self.args.sha256 and self.args.md5) or (self.args.sha1 and self.args.md5):
            raise ValidationException('Please specify only one hashing algorithm')

    def _check_number_is_positive(self):
        # None means download all
        if not (self.args.number is None or self.args.number > 0):
            raise ValidationException('Number of apks must be positive')






from modules.entities.entity import Entity
from modules.entities.value import DateBoundedValue, IntBoundedValue


class Criteria(Entity):

    def __init__(self, dex_date_from=None, dex_date_to=None, apk_size_from=None, apk_size_to=None, vt_detection_from=None, vt_detection_to=None, markets=None, pkg_name=None, sha256=None, sha1=None, md5=None):
        self.dex_date = DateBoundedValue(dex_date_from, dex_date_to)
        self.apk_size = IntBoundedValue(apk_size_from, apk_size_to)
        self.vt_detection = IntBoundedValue(vt_detection_from, vt_detection_to)
        self.markets = set(markets) if markets else None
        self.pkg_name = set(pkg_name) if pkg_name else None
        self.sha256 = set(sha256) if sha256 else None
        self.sha1 = set(sha1) if sha1 else None
        self.md5 = set(md5) if md5 else None

    def _key(self):
        return self.dex_date, self.apk_size, self.vt_detection, self.markets, self.pkg_name, self.sha256, self.sha1, self.md5

    def __str__(self):
        return f'dex_date={self.dex_date}, apk_size={self.apk_size}, vt_detection={self.vt_detection}, markets={self.markets}, pkg_name={self.pkg_name}, sha256={self.sha256}, sha1={self.sha1}, md5={self.md5}'




import os
import sys

import click

from modules import az, logging_util
from modules.cli.args import Arguments
from modules.cli.parser import Parser
from modules.cli.validator import Validator
from modules.entities.criteria import Criteria
from modules.exceptions import AzException, NoArgsException



@click.command()
@click.option('--number', '-n', type=click.INT, help='Number of apks to download.')
@click.option('--dexdate', '-d', help='The date on a dex file, format %Y-%m-%d, e.g. 2015-10-03.')
@click.option('--apksize', '-s', help='Apk size, in bytes.')
@click.option('--vtdetection', '-vt', help='Virus total rating, integer.')
@click.option('--pkgname', '-pn', help='Package names.')
@click.option('--markets', '-m', help='Markets, e.g. play.google.com. Possible values (can differ, since repository is updating): 1mobile,angeeks,anzhi,apk_bang,appchina,fdroid,freewarelovers,genome,hiapk,markets,mi.com,play.google.com,proandroid,slideme,torrents.')
@click.option('--sha256', help='SHA256 hashes of apks to download.')
@click.option('--sha1', help='SHA1 hashes of apks to download.')
@click.option('--md5', help='MD5 hashes of apks to download.')
@click.option('--metadata', '-md', help='Metadata. This is a subset of latest.csv column names to keep in metadata.csv. By default sha256,pkg_name,apk_size,dex_date,markets.')
@click.option('--out', '-o', help='Output folder name. By default current directory.')
@click.option('--seed', '-sd', type=click.INT, help='Seed for a random algorithm.')
@click.option('--key', '-k', help='Androzoo api key.')
@click.option('--input-file', '-i', help='Path to input csv.')
@click.option('--threads', '-t', type=click.INT, default=4, help='Number of threads for concurrent download. 4 by default.')
@click.version_option(message='%(version)s')
def run(number, dexdate, apksize, vtdetection, pkgname, markets, metadata, out, seed, sha256, sha1, md5, key, input_file, threads):
    """Downloads specified number of apks satisfying specified criteria from androzoo repository. Saves specified metadata to metadata.csv.
    dexdate, apksize and vtdetection require specifying lower and upper bounds in format lower:upper, both inclusive. One of the bounds can be omitted (i.e. you can write :upper or lower:)
    pkgname, markets, metadata, sha256, sha1, md5 can be either single values or comma separated lists.
    Key and input file can be specified as options or via local or global config file.
    Allows downloading in the multiple threads.

    Sample usage:

    az -n 10 -d 2015-12-11: -s :3000000  -m play.google.com,appchina

    This means: download 10 apks with the dexdate starting from the 2015-12-11(inclusive), size up to 3000000 bytes(inclusive) and present on either play.google.com or appchina
     """

    try:
        args = Arguments(number, dexdate, apksize, vtdetection, markets, pkgname, metadata, sha256, sha1, md5, key, input_file)
        Validator(args).validate()
        logging_util.setup_logging()
        number, *criteria_args, metadata, key, input_file = Parser(args).parse()
        criteria = Criteria(*criteria_args)
        az.run(input_file, key, number, criteria, out_dir=out if out else os.getcwd(), metadata=metadata, seed=seed, threads=threads)
    except NoArgsException:
        with click.Context(run) as ctx:
            click.echo(run.get_help(ctx))
    except AzException as e:
        sys.exit(str(e))
    else:
        sys.exit(0)


if __name__ == '__main__':
    run(['-vt', '0:0'])



from modules.cli.user_config import UserConfig
from modules.enums import DownloadType


class Parser:
    RANGE_ARGS_DELIMITER = ':'
    LIST_ARGS_DELIMITER = ','
    METADATA_DEFAULT_VALUE = ['sha256', 'pkg_name', 'apk_size', 'dex_date', 'markets']

    def __init__(self, args):
        self.args = args

    def parse(self):
        number = int(self.args.number) if self.args.number else DownloadType.ALL
        dex_date_from, dex_date_to = self.args.dexdate.split(self.RANGE_ARGS_DELIMITER) if self.args.dexdate else (None, None)
        apksize_from, apksize_to = self.args.apksize.split(self.RANGE_ARGS_DELIMITER) if self.args.apksize else (None, None)
        vt_detection_from, vt_detection_to = self.args.vtdetection.split(self.RANGE_ARGS_DELIMITER) if self.args.vtdetection else (None, None)
        markets = self.args.markets.split(self.LIST_ARGS_DELIMITER) if self.args.markets else None
        pkg_name = self.args.pkgname.split(self.LIST_ARGS_DELIMITER) if self.args.pkgname else None
        sha256 = self.get_hash_list(self.args.sha256) if self.args.sha256 else None
        sha1 = self.get_hash_list(self.args.sha1) if self.args.sha1 else None
        md5 = self.get_hash_list(self.args.md5) if self.args.md5 else None
        metadata = self.args.metadata.split(self.LIST_ARGS_DELIMITER) if self.args.metadata else self.METADATA_DEFAULT_VALUE
        key, input_file = self.args.key, self.args.input_file
        if not key or not input_file:
            user_config = UserConfig(key, input_file)
            input_file = input_file if input_file else user_config.in_file
            key = key if key else user_config.key

        return number, dex_date_from, dex_date_to, apksize_from, apksize_to, vt_detection_from, vt_detection_to, markets, pkg_name, sha256, sha1, md5, metadata, key, input_file

    def get_hash_list(self, apk_hashes):
        return [apk_hash.upper() for apk_hash in apk_hashes.split(self.LIST_ARGS_DELIMITER)]


from datetime import datetime


class ApkEvaluator:

    DATE_PATTERN = '%Y-%m-%d %H:%M:%S'

    def __init__(self, criteria=None):
        self.criteria = criteria

    def satisfies(self, apk):
        if not apk:
            return False
        return self.satisfies_sha256(apk) and self.satisfies_sha1(apk) and self.satisfies_md5(apk) and self.satisfies_markets(apk) and self.satisfies_size(apk)\
            and self.satisfies_date(apk) and self.satisfies_name(apk) and self.satisfies_vt_detection(apk)

    def satisfies_markets(self, apk):
        satisfies = True
        if self.criteria.markets:
            satisfies = apk.markets and self.criteria.markets.intersection(apk.markets)
        return satisfies

    def satisfies_sha256(self, apk):
        satisfies = True
        if self.criteria.sha256:
            satisfies = apk.sha256 in self.criteria.sha256
        return satisfies

    def satisfies_sha1(self, apk):
        satisfies = True
        if self.criteria.sha1:
            satisfies = apk.sha1 in self.criteria.sha1
        return satisfies

    def satisfies_md5(self, apk):
        satisfies = True
        if self.criteria.md5:
            satisfies = apk.md5 in self.criteria.md5
        return satisfies

    def satisfies_vt_detection(self, apk):
        satisfies = True
        if self.criteria.vt_detection:
            satisfies = self.satisfies_bounds(apk.vt_detection, self.criteria.vt_detection)
        return satisfies

    def satisfies_name(self, apk):
        satisfies = True
        if self.criteria.pkg_name:
            satisfies = apk.pkg_name in self.criteria.pkg_name
        return satisfies

    def satisfies_size(self, apk):
        satisfies = True
        if self.criteria.apk_size:
            satisfies = self.satisfies_bounds(apk.apk_size, self.criteria.apk_size)
        return satisfies

    def satisfies_date(self, apk):
        satisfies = True
        apk_date = datetime.strptime(apk.dex_date, self.DATE_PATTERN) if apk.dex_date else None
        if self.criteria.dex_date:
            satisfies = self.satisfies_bounds(apk_date, self.criteria.dex_date)

        return satisfies

    @staticmethod
    def satisfies_bounds(value, bounded_value):
        satisfies = True
        if value is not None:
            if bounded_value.lower is not None:
                satisfies = satisfies and bounded_value.lower <= value
            if bounded_value.upper is not None:
                satisfies = satisfies and value <= bounded_value.upper
        else:
            satisfies = False
        return satisfies



from modules.entities.entity import Entity


class Dataset(Entity):
    def __init__(self, *apks):
        if apks is None:
            apks = []
        self.apks = list(apks)

    def _key(self):
        return frozenset(self.apks)

    def add(self, apk):
        self.apks.append(apk)

    def __len__(self):
        return len(self.apks)

    def __iter__(self):
        return iter(self.apks)

    def contains(self, subset):
        return set(subset.apks).issubset(set(self.apks))

    def is_empty(self):
        return len(self.apks) == 0



import logging
import os
import sys

import requests

from modules.services.url_constructor import UrlConstructor
from multiprocessing.dummy import Pool


class DatasetDownloader:
    def __init__(self, dataset, threads, url_constructor=None, key='', out_dir='azoo_dataset'):
        self.dataset = dataset
        self.out_dir = out_dir
        if not url_constructor:
            url_constructor = UrlConstructor(key)
        self.url_constructor = url_constructor
        self.threads = threads

    def download(self):
        logging.info(f'DOWNLOADING {len(self.dataset)}, number of threads {self.threads}')
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        with Pool(self.threads) as pool:
            pool.map(self.download_apk, self.dataset)

    def download_apk(self, apk):
        apk_save_path = os.path.join(self.out_dir, apk.pkg_name) + '.apk'
        try:
            if os.path.exists(apk_save_path):
                apk_save_path = apk_save_path.replace('.apk', f'{apk.sha1}.apk')
                logging.warning(f'apk with pkg {apk.pkg_name} already exists, saving by {apk_save_path}')
            logging.debug(f'DOWNLOAD {apk.pkg_name}... ')
            apk_url = self.url_constructor.construct(apk)
            response = requests.get(apk_url)
            with open(apk_save_path, 'wb') as out_file:
                code = response.status_code
                if code != 200:
                    logging.warning(f'HTTP code for {apk.pkg_name} is {code}')
                out_file.write(response.content)
        except:
            logging.error(f'Unexpected error while downloading {apk.pkg_name}: {sys.exc_info()[1]}')



from modules.entities.apk import Apk
from modules.entities.entity import Entity


class Source(Entity):
    def __init__(self, input_file=None):
        self.input_file = open(input_file)
        next(self.input_file) # skip header

    def __iter__(self):
        return self

    def __next__(self):
        try:
            line = next(self.input_file)
        except StopIteration:
            self.input_file.close()
            raise StopIteration
        columns = line.strip('\n').replace('"', '').split(',')
        return Apk(*columns)

    def _key(self):
        pass




class UrlConstructor:
    def __init__(self, key='', base_url='https://androzoo.uni.lu/api/download?apikey={0}&sha256={01}'):
        self.base_url = base_url
        self.key = key

    def construct(self, apk):
        return self.base_url.format(self.key, apk.sha256)



class AzException(Exception):
    pass


class EmptyDatasetException(AzException):
    pass


class AbsentUserConfigException(AzException):
    pass


class NotDefinedConfigValue(AzException):
    pass


class ValidationException(AzException):
    pass

class NoArgsException(AzException):
    pass



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='azoo',
    version='1.3.0',
    author="Artsiom Kushnariou",
    author_email="kushnerovartem@gmail.com",
    description="Downloads apks from androzoo repository https://androzoo.uni.lu/",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ArtemKushnerov/az',
    entry_points={
        'console_scripts': [
            'az=modules.main:run'
        ]
    },
    packages=find_packages(),
    package_data={
        '': ['*.yaml'],
    },
    install_requires=[
        'click==6.7',
        'setuptools==41.0.1',
        'python_dateutil==2.7.2',
        'PyYAML==4.2b1',
        'requests==2.20.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from dateutil.parser import parse

from modules.entities.entity import Entity


class BoundedValue(Entity):
    def __init__(self, lower, upper, preprocess=None):
        if preprocess:
            lower = preprocess(lower) if lower else None
            upper = preprocess(upper) if upper else None
        self.lower = lower
        self.upper = upper

    def _key(self):
        return self.lower, self.upper

    def __bool__(self):
        return True if self.lower is not None or self.upper is not None else False


class DateBoundedValue(BoundedValue):
    def __init__(self, lower, upper):
        super().__init__(lower, upper, preprocess=parse)


class IntBoundedValue(BoundedValue):
    def __init__(self, lower, upper):
        super().__init__(lower, upper, preprocess=int)




import random

from modules.entities.dataset.dataset import Dataset
from modules.enums import DownloadType


class RandomPicker:
    def __init__(self, seed=None):
        if seed is not None:
            random.seed(seed)

    def get_random_subset(self, input_dataset, requested_size):
        result = []
        for num, apk in enumerate(input_dataset, 1):
            if requested_size is DownloadType.ALL or len(result) < requested_size:
                result.append(apk)
            else:
                s = int(random.random() * num)
                if s < requested_size:
                    result[s] = apk
        return Dataset(*result)



import logging.config
import sys

from modules.entities.dataset.filtered_dataset import FilteredDataset
from modules.entities.source import Source
from modules.services.dataset_downloader import DatasetDownloader
from modules.services.metadata_saver import MetadataSaver
from modules.services.random_picker import RandomPicker


def run(input_file, key, number, criteria, metadata, threads, out_dir='azoo_dataset', seed=None, ):
    logging.info(f'START. APKS TO DOWNLOAD: {str(number)}')
    source = None
    try:
        source = Source(input_file=input_file)

        dataset = FilteredDataset(source, criteria)
        random_subset = RandomPicker(seed=seed).get_random_subset(dataset, number)
        DatasetDownloader(dataset=random_subset, out_dir=out_dir, key=key, threads=threads).download()

        MetadataSaver(random_subset, out_dir).save(metadata)
    except KeyboardInterrupt:
        logging.info('Keyboard interrupt')
        sys.exit()
    finally:
        if source:
            source.input_file.close()




import logging
import os


class MetadataSaver:
    def __init__(self, dataset, out_dir):
        self.dataset = dataset
        self.out_dir = out_dir
        self.file_name = 'metadata.csv'

    def save(self, metadata):
        logging.info(f'SAVE METADATA {metadata}')

        data = ','.join(metadata) + '\n'
        for apk in self.dataset:
            data += self.get_line(apk, metadata) + '\n'

        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        save_path = os.path.join(self.out_dir, self.file_name)
        with open(save_path, 'w+') as file:
            file.write(data)
            file.flush()

    def get_line(self, apk, fields):
        line = ''
        for field in fields:
            line += str(getattr(apk, field)) + ','
        line = line[:-1]
        return line



import logging.config
from pkg_resources import resource_stream

import yaml


def setup_logging():
    with resource_stream('modules.resources', 'logging.yaml') as f:
        logging.config.dictConfig(yaml.safe_load(f.read()))
